
from lxml import etree
import pandas as pd
import os
from tqdm import tqdm
import time
from .grabMzmlFunctions import grabMzmlData
from .grabMzxmlFunctions import grabMzxmlData

def grabMSdata(files, grab_what="everything", verbosity=None):
    """Grab mass-spectrometry data from file(s)

    The main `pylgrams` function. This function accepts a list of the files that will
    be read into working memory and returns a list of `data.table`s
    containing the requested information. What information is requested is
    determined by the `grab_what` argument, which can include MS1 or MS2
     information. This function serves as a wrapper around both
    `grabMzmlData` and `grabMzxmlData` and handles multiple files, but those two
    have also been exposed to the user in case super-simple handling is desired.
    Retention times are reported in minutes, and will be converted automatically
    if they are encoded in seconds.

    Parameters
    ----------
    files: A character vector of filenames to read into memory.
    Both absolute and relative paths are acceptable.

    grab_what: What data should be read from the file? Options include
    "MS1" for data only from the first spectrometer, "MS2" for fragmentation
    data, "BPC" for a base peak chromatogram, or "TIC" for a total ion chromatogram.
    These options can be combined (i.e. `grab_data=["MS1", "MS2"]`) or
    this argument can be set to "everything" to extract all of the above.

    verbosity: Three levels of processing output to the console are
    available, with increasing verbosity corresponding to higher integers. A
    verbosity of zero means that no output will be produced, useful when
    wrapping within larger functions. A verbosity of 1 will produce a progress
    bar that updates after each file is read. A verbosity of 2 or higher will
    produce timing output for each individual file read in. The default, NULL,
    will select between 1 and 2 depending on the number of files being read: if
    a single file, verbosity is set to 2; if multiple files, verbosity is set
    to 1.

    Returns
    -------
    A list of Pandas data frames, each named after the arguments requested in
    grab_what. E.g. ["MS1"] contains MS1 information, ["MS2"] contains fragmentation
    info, etc. MS1 data has four columns: retention time (rt), mass-to-charge
    (mz), intensity (int), and filename. MS2 data has six: retention time (rt),
    precursor m/z (premz), fragment m/z (fragmz), fragment intensity (int),
    collision energy (voltage), and filename. Data
    requested that does not exist in the provided files (such as MS2 data
    requested from MS1-only files) will return an empty (length zero)
    data frames. The data frames extracted from each of the individual files are
    collected into one large table.

    """
    if not files:
        raise ValueError("No files provided")
    if isinstance(files, str):
        files=[files]
    
    if grab_what=="everything":
        grab_what = ["MS1", "MS2", "BPC", "TIC"]

    if verbosity is None:
        verbosity = 2 if len(files) == 1 else 1
    
    all_file_data = {}
    if verbosity > 0:
        if len(files) >= 2:
            pb = tqdm(total=len(files))
        start_time = time.time()
    
    for file in files:
        filename = os.path.basename(file)
        if '.mzml' in filename.lower():
            out_data = grabMzmlData(filename=file, grab_what=grab_what, verbosity=verbosity)
        elif '.mzxml' in filename.lower():
            out_data = grabMzxmlData(filename=file, grab_what=grab_what, verbosity=verbosity)
        else:
            raise ValueError(f"Unable to determine file type for {filename}")
        
        out_data["MS1"]["filename"] = filename
        out_data["MS2"]["filename"] = filename
        out_data["BPC"]["filename"] = filename
        out_data["TIC"]["filename"] = filename
        all_file_data[filename] = out_data
        
        if verbosity > 0 and len(files) >= 2:
            pb.update(1)

    all_file_data_output = {}
    all_file_ms1 = [file_data['MS1'] for file_data in all_file_data.values() if 'MS1' in file_data]
    all_file_data_output["MS1"] = pd.concat(all_file_ms1, ignore_index=True)
    all_file_ms2 = [file_data['MS2'] for file_data in all_file_data.values() if 'MS2' in file_data]
    all_file_data_output["MS2"] = pd.concat(all_file_ms2, ignore_index=True)
    all_file_bpc = [file_data['BPC'] for file_data in all_file_data.values() if 'BPC' in file_data]
    all_file_data_output["BPC"] = pd.concat(all_file_bpc, ignore_index=True)
    all_file_tic = [file_data['TIC'] for file_data in all_file_data.values() if 'TIC' in file_data]
    all_file_data_output["TIC"] = pd.concat(all_file_tic, ignore_index=True)
    if verbosity > 0:
        if len(files) >= 2:
            pb.close()
        time_total = round(time.time() - start_time, 2)
        print("Total time:", time_total, "seconds")
    return all_file_data_output

__all__ = ["grabMSdata", "grabMzmlData", "grabMzxmlData"]



from lxml import etree
import zlib
import base64
import struct
import math
import numpy as np
import pandas as pd
import os
import time

def grabMzmlData(filename, grab_what, verbosity=0):
    """Get mass-spectrometry data from an mzML file

    This function handles the mzML side of things, reading in files that are
    written in the mzML format. Much of the code is similar to the mzXML format,
    but the xpath handles are different and the mz/int array is encoded as two
    separate entries rather than simultaneously. This function has been exposed
    to the user in case per-file optimization (such as peakpicking or additional
    filtering) is desired before the full data object is returned.

    Parameters
    ----------
    files: A character vector of filenames to read into memory.
    Both absolute and relative paths are acceptable.

    grab_what: What data should be read from the file? Options include
    "MS1" for data only from the first spectrometer, "MS2" for fragmentation
    data, "BPC" for a base peak chromatogram, or "TIC" for a total ion chromatogram.
    These options can be combined (i.e. `grab_data=["MS1", "MS2"]`) or
    this argument can be set to "everything" to extract all of the above.

    verbosity: Three levels of processing output to the console are
    available, with increasing verbosity corresponding to higher integers. A
    verbosity of zero means that no output will be produced, useful when
    wrapping within larger functions. A verbosity of 1 will produce a progress
    bar that updates after each file is read. A verbosity of 2 or higher will
    produce timing output for each individual file read in. The default, NULL,
    will select between 1 and 2 depending on the number of files being read: if
    a single file, verbosity is set to 2; if multiple files, verbosity is set
    to 1.

    Returns
    -------
    A list of Pandas data frames, each named after the arguments requested in
    grab_what. E.g. ["MS1"] contains MS1 information, ["MS2"] contains fragmentation
    info, etc. MS1 data has four columns: retention time (rt), mass-to-charge
    (mz), intensity (int), and filename. MS2 data has six: retention time (rt),
    precursor m/z (premz), fragment m/z (fragmz), fragment intensity (int),
    collision energy (voltage), and filename. Data
    requested that does not exist in the provided files (such as MS2 data
    requested from MS1-only files) will return an empty (length zero)
    data frame.

    """
    if verbosity > 1:
        print(f"\nReading file {os.path.basename(filename)}... ")
        last_time = time.time()

    xml_data = etree.parse(filename)
    output_data = {}
    file_metadata = grabMzmlEncodingData(xml_data)

    if "MS1" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading MS1 data...")
        output_data["MS1"] = grabMzmlMS1(xml_data, file_metadata)

    if "MS2" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading MS2 data...")
        output_data["MS2"] = grabMzmlMS2(xml_data, file_metadata)

    if "BPC" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading BPC data...")
        output_data["BPC"] = grabMzmlBPC(xml_data)

    if "TIC" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading TIC data...")
        output_data["TIC"] = grabMzmlBPC(xml_data, TIC=True)

    if verbosity > 1:
        time_total = round(time.time() - last_time, 2)
        print("Total time:", time_total, "seconds")

    return output_data

def grabMzmlBPC(xml_data, TIC=False):
    if TIC == True:
        cvparam_name = "total ion current"
    else:
        cvparam_name = "base peak intensity"

    ms1_xpath = "//d1:spectrum[d1:cvParam[@name=\"ms level\" and @value=\"1\"]][d1:cvParam[@name=\"" + cvparam_name + "\"]]"
    ms1_nodes = xml_data.xpath(ms1_xpath, namespaces={'d1': 'http://psi.hupo.org/ms/mzml'})
    rt_vals = grabSpectraRt(ms1_nodes)

    int_xpath = 'd1:cvParam[@name="' + cvparam_name + '"]'
    int_nodes = [node.xpath(int_xpath, namespaces={'d1': 'http://psi.hupo.org/ms/mzml'}) for node in ms1_nodes]
    int_nodes = [item for sublist in int_nodes for item in sublist]
    int_vals = [float(node.attrib["value"]) for node in int_nodes]

    all_data = pd.DataFrame({
        'rt': rt_vals,
        'int': int_vals
    })
    return(all_data)


def grabMzmlEncodingData(xml_data):
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}

    init_xpath = "//*[self::d1:spectrum or self::d1:chromatogram]"
    init_node = xml_data.xpath(init_xpath, namespaces=ns)
    if not init_node:
        raise ValueError("Unable to find a spectrum or chromatogram node from which to extract metadata")

    init_node = init_node[0]

    compr_xpath = '//d1:cvParam[@accession="MS:1000574" or @accession="MS:1000576"]'
    compr_node = init_node.xpath(compr_xpath, namespaces=ns)[0]
    compr_type = compr_node.attrib["name"]

    compr = {
        "zlib": "gzip",
        "zlib compression": "gzip",
        "no compression": "none",
        "none": "none"
    }[compr_type]

    mz_precision_xpath = '//d1:cvParam[@accession="MS:1000523"]'
    mz_bit_node = init_node.xpath(mz_precision_xpath, namespaces=ns)[0]
    mz_bit_type = mz_bit_node.attrib["name"]
    mz_precision = int(mz_bit_type.split("-")[0]) / 8

    int_bit_xpath = '//d1:cvParam[@accession="MS:1000521"]'
    int_bit_node = init_node.xpath(int_bit_xpath, namespaces=ns)
    if int_bit_node:
        int_bit_node = int_bit_node[0]
        int_bit_type = int_bit_node.attrib["name"]
        int_precision = int(int_bit_type.split("-")[0]) / 8
    else:
        int_precision = mz_precision

    # Ensure mz_precision and int_precision are not NaN
    mz_precision = mz_precision if not math.isnan(mz_precision) else int_precision
    int_precision = int_precision if not math.isnan(int_precision) else mz_precision

    return {
        "compression": compr,
        "mz_precision": int(mz_precision),
        "int_precision": int(int_precision),
        "endi_enc": "little"
    }

def grabMzmlMS1(xml_data, file_metadata):
    ms1_xpath = "//d1:spectrum[d1:cvParam[@name=\"ms level\" and @value=\"1\"]]"
    ms1_nodes = xml_data.xpath(ms1_xpath, namespaces={'d1': 'http://psi.hupo.org/ms/mzml'})
    
    if not ms1_nodes:
        return pd.DataFrame({'rt': [], 'mz': [], 'int': []})
    
    rt_vals = grabSpectraRt(ms1_nodes)
    mz_vals = grabSpectraMz(ms1_nodes, file_metadata)
    int_vals = grabSpectraInt(ms1_nodes, file_metadata)
    
    all_data = pd.DataFrame({
        'rt': np.repeat(rt_vals, [len(x) for x in mz_vals]),
        'mz': np.concatenate(mz_vals),
        'int': np.concatenate(int_vals)
    })
    
    return all_data

def grabMzmlMS2(xml_data, file_metadata):
    ms2_xpath = "//d1:spectrum[d1:cvParam[@name=\"ms level\" and @value=\"2\"]]"
    ms2_nodes = xml_data.xpath(ms2_xpath, namespaces={'d1': 'http://psi.hupo.org/ms/mzml'})
    
    if not ms2_nodes:
        return pd.DataFrame({'rt': [], 'premz': [], 'fragmz': [], 'int': [], 'voltage': []})
    
    rt_vals = grabSpectraRt(ms2_nodes)
    premz_vals = grabSpectraPremz(ms2_nodes)
    voltage = grabSpectraVoltage(ms2_nodes)
    mz_vals = grabSpectraMz(ms2_nodes, file_metadata)
    int_vals = grabSpectraInt(ms2_nodes, file_metadata)
    
    all_data = {
        'rt': np.concatenate([[rt_val] * len(mz_val) for rt_val, mz_val in zip(rt_vals, mz_vals)]),
        'premz': np.concatenate([[premz_val] * len(mz_val) for premz_val, mz_val in zip(premz_vals, mz_vals)]),
        'fragmz': np.concatenate(mz_vals),
        'int': np.concatenate(int_vals),
        'voltage': np.concatenate([[voltage_val] * len(mz_val) for voltage_val, mz_val in zip(voltage, mz_vals)])
    }
    
    return pd.DataFrame(all_data)

def grabSpectraRt(xml_nodes):
    rt_xpath = './/d1:scanList/d1:scan/d1:cvParam[@name="scan start time"]'
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}
    
    # Find all the relevant nodes
    rt_nodes = [node.xpath(rt_xpath, namespaces=ns) for node in xml_nodes]
    rt_nodes = [item for sublist in rt_nodes for item in sublist]
    
    # Extract the unit names and values
    rt_units = set(node.attrib.get('unitName') for node in rt_nodes)
    rt_vals = [float(node.attrib['value']) for node in rt_nodes]

    # Convert to minutes if not already in minutes
    if "minute" not in rt_units:
        rt_vals = [val / 60 for val in rt_vals]

    return rt_vals

def grabSpectraMz(xml_nodes, file_metadata):
    mz_xpath = './/d1:binaryDataArrayList/d1:binaryDataArray[1]/d1:binary'
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}
    
    mz_vals = [node.xpath(mz_xpath, namespaces=ns) for node in xml_nodes]
    mz_vals = [item for sublist in mz_vals for item in sublist]  # Flatten the list
    mz_vals = [node.text for node in mz_vals]
    
    result = []
    for binary in mz_vals:
        if not binary:
            result.append([])
            continue
        decoded_binary = base64.b64decode(binary)
        raw_binary = bytes(decoded_binary)
        if file_metadata['compression'] == 'none':
            decomp_binary = raw_binary
        elif file_metadata['compression'] == 'zlib':  # Assuming zlib compression
            decomp_binary = zlib.decompress(raw_binary)
        elif file_metadata['compression'] == 'gzip':
            decomp_binary = zlib.decompress(raw_binary)
        else:
            raise ValueError(f"Unsupported compression type: {file_metadata['compression']}")
        # Read the binary data as doubles (8 bytes for each double value)
        num_doubles = len(decomp_binary) // file_metadata['mz_precision']
        double_format = '{}d'.format(int(num_doubles))
        unpacked_data = struct.unpack(double_format, decomp_binary)
        result.append(unpacked_data)
    return result

def grabSpectraInt(xml_nodes, file_metadata):
    int_xpath = './/d1:binaryDataArrayList/d1:binaryDataArray[2]/d1:binary'
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}
    
    int_vals = [node.xpath(int_xpath, namespaces=ns) for node in xml_nodes]
    int_vals = [item for sublist in int_vals for item in sublist]  # Flatten the list
    int_vals = [node.text for node in int_vals]
    
    result = []
    for binary in int_vals:
        if not binary:
            result.append([])
            continue
        decoded_binary = base64.b64decode(binary)
        raw_binary = bytes(decoded_binary)
        if file_metadata['compression'] == 'none':
            decomp_binary = raw_binary
        elif file_metadata['compression'] == 'zlib':
            decomp_binary = zlib.decompress(raw_binary)
        elif file_metadata['compression'] == 'gzip':
            decomp_binary = zlib.decompress(raw_binary)
        else:
            raise ValueError(f"Unsupported compression type: {file_metadata['compression']}")
        num_doubles = len(decomp_binary) // file_metadata['int_precision']
        double_format = '{}f'.format(int(num_doubles))
        unpacked_data = struct.unpack(double_format, decomp_binary)
        result.append(unpacked_data)
    return result

def grabSpectraVoltage(xml_nodes):
    volt_xpath = "//d1:precursorList/d1:precursor/d1:activation/d1:cvParam[@name=\"collision energy\"]"
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}
    volt_nodes = [node.xpath(volt_xpath, namespaces=ns) for node in xml_nodes]
    volt_nodes = [item for sublist in volt_nodes for item in sublist]
    if not volt_nodes:
        return [None] * len(xml_nodes)
    volt_vals = [int(node.get("value")) for node in volt_nodes]
    return(volt_vals)

def grabSpectraPremz(xml_nodes):
    premz_xpath = "//d1:precursorList/d1:precursor/d1:selectedIonList/d1:selectedIon/d1:cvParam[@name=\"selected ion m/z\"]"
    ns = {'d1': 'http://psi.hupo.org/ms/mzml'}
    premz_nodes = [node.xpath(premz_xpath, namespaces=ns) for node in xml_nodes]
    premz_nodes = [item for sublist in premz_nodes for item in sublist]
    premz_vals = [float(node.get("value")) for node in premz_nodes]
    return(premz_vals)

def timeReport(last_time, text):
    current_time = time.time()
    time_total = round(current_time - last_time, 2)
    print(text, time_total, "seconds")
    return current_time

__all__ = ["grabMzmlData"]




import pytest
import pandas as pd
import os
from pylgrams import grabMSdata

sample_files = ["src/pylgrams/example_data/LB12HL_AB.mzML.gz",
     "src/pylgrams/example_data/LB12HL_AB.mzXML.gz",
     "src/pylgrams/example_data/LB12HL_CD.mzML.gz",
     "src/pylgrams/example_data/LB12HL_EF.mzML.gz",
     "src/pylgrams/example_data/S30657.mzML.gz",
     "src/pylgrams/example_data/S30657.mzXML.gz"]

msdata = grabMSdata(sample_files, grab_what=["MS1", "MS2", "BPC", "TIC"])

def test_msdata_structure():
    assert isinstance(msdata, dict), "msdata should be a dictionary"
    assert len(msdata) == 4, "msdata should be of length 4"
    assert msdata.keys() == {"MS1", "MS2", "BPC", "TIC"}, "msdata is named incorrectly"

def test_ms1_structure():
    ms1_data = msdata["MS1"]
    assert isinstance(ms1_data, pd.DataFrame), "MS1 is not a pd.DataFrame"
    assert list(ms1_data) == ["rt", "mz", "int", "filename"], "MS1 is named incorrectly"
    assert len(ms1_data["rt"]) > 0, "No rows present in MS1"
    assert all(ms1_data.drop("filename", axis=1) > 0), "Negative values present in MS1"


def test_ms2_structure():
    ms2_data = msdata["MS2"]
    assert isinstance(ms2_data, pd.DataFrame), "MS2 is not a pd.DataFrame"
    assert list(ms2_data) == ["rt", "premz", "fragmz", "int", "voltage", "filename"], "MS2 is named incorrectly"
    assert len(ms2_data["rt"]) > 0, "No rows present in MS2"
    assert all(ms2_data.drop("filename", axis=1) > 0), "Negative values present in MS2"

def test_bpctic_structure():
    bpc_data = msdata["BPC"]
    assert isinstance(bpc_data, pd.DataFrame), "BPC is not a pd.DataFrame"
    assert list(bpc_data) == ["rt", "int", "filename"], "BPC is named incorrectly"
    assert len(bpc_data["rt"]) > 0, "No rows present in BPC"
    assert all(bpc_data.drop("filename", axis=1) > 0), "Negative values present in BPC"

    tic_data = msdata["TIC"]
    assert isinstance(tic_data, pd.DataFrame), "TIC is not a pd.DataFrame"
    assert list(tic_data) == ["rt", "int", "filename"], "TIC is named incorrectly"
    assert len(tic_data["rt"]) > 0, "No rows present in TIC"
    assert all(tic_data.drop("filename", axis=1) > 0), "Negative values present in TIC"

def test_multifile_things():
    assert msdata["MS1"]["filename"].unique().tolist() == [os.path.basename(filename) for filename in sample_files]
    assert len(msdata["MS1"]["filename"].unique()) == len(sample_files)


if __name__ == "__main__":
    pytest.main()




from lxml import etree
import zlib
import base64
import struct
import math
import numpy as np
import pandas as pd
import os
import time
import re

def grabMzxmlData(filename, grab_what, verbosity=0):
    """Get mass-spectrometry data from an mzXML file

    This function handles the mzXML side of things, reading in files that are
    written in the mzXML format. Much of the code is similar to the mzML format,
    but the xpath handles are different and the mz/int array is encoded as a
    single entry rather than two separate. This function has been exposed
    to the user in case per-file optimization (such as peakpicking or additional
    filtering) is desired before the full data object is returned.

    Parameters
    ----------
    files: A character vector of filenames to read into memory.
    Both absolute and relative paths are acceptable.

    grab_what: What data should be read from the file? Options include
    "MS1" for data only from the first spectrometer, "MS2" for fragmentation
    data, "BPC" for a base peak chromatogram, or "TIC" for a total ion chromatogram.
    These options can be combined (i.e. `grab_data=["MS1", "MS2"]`) or
    this argument can be set to "everything" to extract all of the above.

    verbosity: Three levels of processing output to the console are
    available, with increasing verbosity corresponding to higher integers. A
    verbosity of zero means that no output will be produced, useful when
    wrapping within larger functions. A verbosity of 1 will produce a progress
    bar that updates after each file is read. A verbosity of 2 or higher will
    produce timing output for each individual file read in. The default, NULL,
    will select between 1 and 2 depending on the number of files being read: if
    a single file, verbosity is set to 2; if multiple files, verbosity is set
    to 1.

    Returns
    -------
    A list of Pandas data frames, each named after the arguments requested in
    grab_what. E.g. ["MS1"] contains MS1 information, ["MS2"] contains fragmentation
    info, etc. MS1 data has four columns: retention time (rt), mass-to-charge
    (mz), intensity (int), and filename. MS2 data has six: retention time (rt),
    precursor m/z (premz), fragment m/z (fragmz), fragment intensity (int),
    collision energy (voltage), and filename. Data
    requested that does not exist in the provided files (such as MS2 data
    requested from MS1-only files) will return an empty (length zero)
    data frame.

    """
    if verbosity > 1:
        print(f"\nReading file {os.path.basename(filename)}... ")
        last_time = time.time()

    xml_data = etree.parse(filename)
    output_data = {}
    file_metadata = grabMzxmlEncodingData(xml_data)

    if "MS1" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading MS1 data...")
        output_data["MS1"] = grabMzxmlMS1(xml_data, file_metadata)

    if "MS2" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading MS2 data...")
        output_data["MS2"] = grabMzxmlMS2(xml_data, file_metadata)

    if "BPC" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading BPC data...")
        output_data["BPC"] = grabMzxmlBPC(xml_data)

    if "TIC" in grab_what:
        if verbosity > 1:
            last_time = timeReport(last_time, "Reading TIC data...")
        output_data["TIC"] = grabMzxmlBPC(xml_data, TIC=True)

    if verbosity > 1:
        time_total = round(time.time() - last_time, 2)
        print("Total time:", time_total, "seconds")

    return output_data

def grabMzxmlEncodingData(xml_data):
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    peak_metadata = xml_data.xpath('//d1:peaks', namespaces=ns)[0]

    compression_type = peak_metadata.get("compressionType")
    compression = {
        'zlib': 'gzip',
        'zlib compression': 'gzip',
        'no compression': 'none',
        'none': 'none'
    }.get(compression_type, 'none')

    precision = int(peak_metadata.get("precision")) / 8

    byte_order = peak_metadata.get("byteOrder")
    endianness_encoding = {
        'network': 'big'
    }.get(byte_order, None)

    return {
        'compression': compression,
        'precision': precision,
        'endi_enc': endianness_encoding
    }

def grabMzxmlBPC(xml_data, TIC=False):
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    ms1_xpath = '//d1:scan[@msLevel="1" and @peaksCount>0]'
    ms1_nodes = xml_data.xpath(ms1_xpath, namespaces=ns)

    rt_vals = grabMzxmlSpectraRt(ms1_nodes)

    if TIC == True:
        cvparam_name = "totIonCurrent"
    else:
        cvparam_name = "basePeakIntensity"
    int_vals = [float(node.attrib[cvparam_name]) for node in ms1_nodes]

    all_data = pd.DataFrame({
        'rt': rt_vals,
        'int': int_vals
    })
    return (all_data)

def grabMzxmlMS1(xml_data, file_metadata):
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    ms1_xpath = '//d1:scan[@msLevel="1" and @peaksCount>0]'
    ms1_nodes = xml_data.xpath(ms1_xpath, namespaces=ns)
    
    if not ms1_nodes:
        return pd.DataFrame({'rt': [], 'mz': [], 'int': []})
    
    rt_vals = grabMzxmlSpectraRt(ms1_nodes)
    mzint_vals = grabMzxmlSpectraMzInt(ms1_nodes, file_metadata)

    rt_array = np.array(rt_vals)
    rt_repeat = np.repeat(rt_array, [len(arr) for arr in mzint_vals])
    mzint_flat = np.concatenate(mzint_vals)
    dt_array = np.column_stack((rt_repeat, mzint_flat))
    all_data = pd.DataFrame(dt_array, columns=["rt", "mz", "int"])

    return all_data

def grabMzxmlMS2(xml_data, file_metadata):
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    ms2_xpath = '//d1:scan[@msLevel="2" and @peaksCount>0]'
    ms2_nodes = xml_data.xpath(ms2_xpath, namespaces=ns)
    
    if not ms2_nodes:
        return pd.DataFrame({'rt': [], 'premz': [], 'fragmz': [], 'int': [], 'voltage': []})
    
    rt_vals = grabMzxmlSpectraRt(ms2_nodes)
    premz_vals = grabMzxmlSpectraPremz(ms2_nodes)
    volt_vals = grabMzxmlSpectraVoltage(ms2_nodes)
    mzint_vals = grabMzxmlSpectraMzInt(ms2_nodes, file_metadata)

    rt_repeat = np.repeat(rt_vals, [len(arr) for arr in mzint_vals])
    premz_repeat = np.repeat(premz_vals, [len(arr) for arr in mzint_vals])
    volt_repeat = np.repeat(volt_vals, [len(arr) for arr in mzint_vals])
    mzint_flat = np.concatenate(mzint_vals)
    dt_array = np.column_stack((rt_repeat, premz_repeat, mzint_flat, volt_repeat))
    all_data = pd.DataFrame(dt_array, columns=["rt", "premz", "fragmz", "int", "voltage"])

    return all_data

def grabMzxmlSpectraRt(xml_nodes):
    rt_attrs = [node.get("retentionTime") for node in xml_nodes]
    rt_units = set(re.sub(r".*[0-9]", "", rt) for rt in rt_attrs)
    rt_vals = [float(re.sub(r"PT|S", "", rt)) for rt in rt_attrs]
    if "S" in rt_units:
        rt_vals = [rt / 60 for rt in rt_vals]
    return(rt_vals)

def grabMzxmlSpectraMzInt(xml_nodes, file_metadata):
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    all_peak_nodes = [node.find(".//d1:peaks", namespaces=ns).text for node in xml_nodes]

    vals = []
    for binary in all_peak_nodes:
        if not binary or len(binary) == 0:
            vals.append(np.empty((0, 2)))  # Empty matrix
            continue
        decoded_binary = base64.b64decode(binary)
        if file_metadata['compression'] == 'none':
            decomp_binary = decoded_binary
        elif file_metadata['compression'] == 'zlib':
            decomp_binary = zlib.decompress(decoded_binary)
        else:
            raise ValueError(f"Unsupported compression type: {file_metadata['compression']}")

        if file_metadata['precision'] == 8:
            fmt = 'd'  # Double precision (64-bit)
        else:
            fmt = 'f'  # Single precision (32-bit)
        if file_metadata['endi_enc'] == 'big':
            endi = '>'
        else:
            endi = '<'

        num_values = len(decomp_binary) // struct.calcsize(fmt)
        final_binary = struct.unpack(f'{endi}{num_values}{fmt}', decomp_binary)
        final_binary = np.array(final_binary).reshape(-1, 2)
        vals.append(final_binary)
    return vals

def grabMzxmlSpectraPremz(xml_nodes):
    premz_xpath = "d1:precursorMz"
    ns = {'d1': 'http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'}
    premz_nodes = [node.xpath(premz_xpath, namespaces=ns) for node in xml_nodes]
    premz_nodes = [item for sublist in premz_nodes for item in sublist]
    if not premz_nodes:
        return [None] * len(xml_nodes)
    premz_vals = [float(node.text) for node in premz_nodes]
    return(premz_vals)

def grabMzxmlSpectraVoltage(xml_nodes):
    volt_vals = []
    for node in xml_nodes:
        collision_energy_attr = node.get("collisionEnergy")
        if collision_energy_attr is not None:
            volt_vals.append(float(collision_energy_attr))
        else:
            volt_vals.append(None)
    return(volt_vals)

def timeReport(last_time, text):
    current_time = time.time()
    time_total = round(current_time - last_time, 2)
    print(text, time_total, "seconds")
    return current_time

__all__ = ["grabMzxmlData"]




import pytest
import pandas as pd
from pylgrams import grabMzxmlData

# Sample test files (replace these with actual test file paths)
sample_file = "src/pylgrams/example_data/S30657.mzXML.gz"

@pytest.fixture
def msdata():
    return grabMzxmlData(sample_file, grab_what=["MS1", "MS2", "BPC", "TIC"])

def test_msdata_structure(msdata):
    assert isinstance(msdata, dict), "msdata should be a dictionary"
    assert len(msdata) == 4, "msdata should be of length 4"
    assert msdata.keys() == {"MS1", "MS2", "BPC", "TIC"}, "msdata is named incorrectly"

def test_ms1_structure(msdata):
    ms1_data = msdata["MS1"]
    assert isinstance(ms1_data, pd.DataFrame), "MS1 is not a pd.DataFrame"
    assert list(ms1_data) == ["rt", "mz", "int"], "MS1 is named incorrectly"
    assert len(ms1_data["rt"]) > 0, "No rows present in MS1"
    assert all(ms1_data > 0), "Negative values present in MS1"


def test_ms2_structure(msdata):
    ms2_data = msdata["MS2"]
    assert isinstance(ms2_data, pd.DataFrame), "MS2 is not a pd.DataFrame"
    assert list(ms2_data) == ["rt", "premz", "fragmz", "int", "voltage"], "MS2 is named incorrectly"
    assert len(ms2_data["rt"]) > 0, "No rows present in MS2"
    assert all(ms2_data > 0), "Negative values present in MS2"

def test_bpctic_structure(msdata):
    bpc_data = msdata["BPC"]
    assert isinstance(bpc_data, pd.DataFrame), "BPC is not a pd.DataFrame"
    assert list(bpc_data) == ["rt", "int"], "BPC is named incorrectly"
    assert len(bpc_data["rt"]) > 0, "No rows present in BPC"
    assert all(bpc_data > 0), "Negative values present in BPC"

    tic_data = msdata["TIC"]
    assert isinstance(tic_data, pd.DataFrame), "TIC is not a pd.DataFrame"
    assert list(tic_data) == ["rt", "int"], "TIC is named incorrectly"
    assert len(tic_data["rt"]) > 0, "No rows present in TIC"
    assert all(tic_data > 0), "Negative values present in TIC"

def test_msdata_comp(msdata):
    assert all(msdata["TIC"]["int"] >= msdata["BPC"]["int"])
    assert len(msdata["TIC"]["rt"]) == len(msdata["BPC"]["rt"])
    assert len(msdata["TIC"]["rt"]) <= len(msdata["MS1"]["rt"])

if __name__ == "__main__":
    pytest.main()



import os
import time
import numpy as np
import pandas as pd
import zlib
import base64
import struct
from tqdm import tqdm
from lxml import etree

from .grabMSdataCode import grabMSdata
from .grabMzmlFunctions import grabMzmlData
from .grabMzxmlFunctions import grabMzxmlData

__all__ = [
    'grabMSdata',
    'grabMzmlData',
    'grabMzxmlData'
]